@@ -39,6 +39,7 @@
 #include <gst/gst.h>
 #include <gst/pbutils/encoding-profile.h>
 
+#include "sj-cell-renderer-text.h"
 #include "rb-gst-media-types.h"
 #include "sj-about.h"
 #include "sj-metadata.h"
@@ -57,12 +58,16 @@ gboolean on_delete_event (GtkWidget *widget, GdkEvent *event, gpointer user_data
 
 static void reread_cd (gboolean ignore_no_media);
 static void update_ui_for_album (AlbumDetails *album);
+static gboolean get_action_state_bool (const char *name);
+static void set_action_state (const char *detailed_name);
 
 /* Prototypes for the signal blocking/unblocking in update_ui_for_album */
 G_MODULE_EXPORT void on_title_edit_changed(GtkEditable *widget, gpointer user_data);
 G_MODULE_EXPORT void on_person_edit_changed(GtkEditable *widget, gpointer user_data);
 G_MODULE_EXPORT void on_year_edit_changed(GtkEditable *widget, gpointer user_data);
 G_MODULE_EXPORT void on_disc_number_edit_changed(GtkEditable *widget, gpointer user_data);
+G_MODULE_EXPORT void submit_bar_response_cb (GtkInfoBar *infobar, int response_id, gpointer user_data);
+G_MODULE_EXPORT void on_activate_move_focus (GtkWidget *widget, gpointer data);
 
 GtkBuilder *builder;
 
@@ -72,7 +77,7 @@ SjExtractor *extractor;
 GSettings *sj_settings;
 
 GtkWidget *main_window;
-static GtkWidget *message_area_eventbox;
+static GtkWidget *submit_bar, *submit_label;
 static GtkWidget *title_entry, *artist_entry, *composer_label, *composer_entry, *duration_label, *genre_entry, *year_entry, *disc_number_entry;
 static GtkWidget *entry_table; /* GtkTable containing composer_entry */
 static GtkTreeViewColumn *composer_column; /* Treeview column containing composers */
@@ -81,8 +86,6 @@ static GtkWidget *status_bar;
 GtkListStore *track_store;
 GtkCellRenderer *toggle_renderer, *title_renderer, *artist_renderer, *composer_renderer;
 
-GtkWidget *current_message_area;
-
 char *path_pattern, *file_pattern;
 GFile *base_uri;
 BraseroDrive *drive = NULL;
@@ -102,6 +105,17 @@ static char *device = NULL, **uris = NULL;
 
 static guint debug_flags = 0;
 
+/**
+ * Shared state for cell editing callbacks
+ */
+typedef struct {
+  GtkTreePath* editable_path;
+  char* text;
+  ViewColumn column;
+} CellCbContext;
+
+static CellCbContext cell_editing_context;
+
 #define DEFAULT_PARANOIA 15
 #define RAISE_WINDOW "raise-window"
 #define SOURCE_BUILDER "../data/sound-juicer.ui"
@@ -183,7 +197,7 @@ static void on_eject_activate (GSimpleAction *action, GVariant *parameter, gpoin
 {
   /* first make sure we're not playing */
   stop_playback ();
-
+  set_action_state ("re-read(false)");
   brasero_drive_eject (drive, FALSE, NULL);
 }
 
@@ -197,8 +211,9 @@ G_MODULE_EXPORT gboolean on_delete_event (GtkWidget *widget, GdkEvent *event, gp
                                      GTK_MESSAGE_QUESTION,
                                      GTK_BUTTONS_NONE,
                                      _("You are currently extracting a CD. Do you want to quit now or continue?"));
-    gtk_dialog_add_button (GTK_DIALOG (dialog), "gtk-quit", GTK_RESPONSE_ACCEPT);
+    gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Quit"), GTK_RESPONSE_ACCEPT);
     gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Continue"), GTK_RESPONSE_REJECT);
+    gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_REJECT);
     response = gtk_dialog_run (GTK_DIALOG (dialog));
     gtk_widget_destroy (dialog);
 
@@ -315,118 +330,37 @@ static void number_cell_icon_data_cb (GtkTreeViewColumn *tree_column,
   }
 }
 
-/* Taken from gedit */
 static void
-set_info_bar_text_and_icon (GtkInfoBar  *infobar,
-                            const gchar *icon_name,
-                            const gchar *primary_text,
-                            const gchar *secondary_text,
-                            GtkWidget   *button)
+set_submit_text (const AlbumDetails *album)
 {
-  GtkWidget *content_area;
-  GtkWidget *hbox_content;
-  GtkWidget *image;
-  GtkWidget *vbox;
-  gchar *primary_markup;
-  gchar *secondary_markup;
-  GtkWidget *primary_label;
-  GtkWidget *secondary_label;
-  AtkObject *ally_target;
+  gchar *text;
 
-  ally_target = gtk_widget_get_accessible (button);
-
-  hbox_content = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 8); 
-  gtk_widget_show (hbox_content);
-
-  image = gtk_image_new_from_icon_name (icon_name, GTK_ICON_SIZE_DIALOG);
-  gtk_widget_show (image);
-  gtk_box_pack_start (GTK_BOX (hbox_content), image, FALSE, FALSE, 0);
-  gtk_misc_set_alignment (GTK_MISC (image), 0.5, 0);
-
-  vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 6); 
-  gtk_widget_show (vbox);
-  gtk_box_pack_start (GTK_BOX (hbox_content), vbox, TRUE, TRUE, 0);
-
-  primary_markup = g_markup_printf_escaped ("<b>%s</b>", primary_text);
-  primary_label = gtk_label_new (primary_markup);
-  g_free (primary_markup);
-  gtk_widget_show (primary_label);
-  gtk_box_pack_start (GTK_BOX (vbox), primary_label, TRUE, TRUE, 0);
-  gtk_label_set_use_markup (GTK_LABEL (primary_label), TRUE);
-  gtk_label_set_line_wrap (GTK_LABEL (primary_label), TRUE);
-  gtk_misc_set_alignment (GTK_MISC (primary_label), 0, 0.5);
-  atk_object_add_relationship (ally_target,
-                               ATK_RELATION_LABELLED_BY,
-                               gtk_widget_get_accessible (primary_label));
-
-  if (secondary_text != NULL) {
-    secondary_markup = g_markup_printf_escaped ("<small>%s</small>",
-                                                secondary_text);
-    secondary_label = gtk_label_new (secondary_markup);
-    g_free (secondary_markup);
-    gtk_widget_show (secondary_label);
-    gtk_box_pack_start (GTK_BOX (vbox), secondary_label, TRUE, TRUE, 0);
-    gtk_label_set_use_markup (GTK_LABEL (secondary_label), TRUE);
-    gtk_label_set_line_wrap (GTK_LABEL (secondary_label), TRUE);
-    gtk_misc_set_alignment (GTK_MISC (secondary_label), 0, 0.5);
-    atk_object_add_relationship (ally_target,
-                                 ATK_RELATION_LABELLED_BY,
-                                 gtk_widget_get_accessible (secondary_label));
+  if (g_str_equal (album->title, _("Unknown Title"))) {
+    text = g_strdup (_("This album is not in the MusicBrainz database."));
+  } else {
+    text = g_strdup_printf (_("Could not find %s by %s on MusicBrainz."),
+                                    album->title, album->artist);
   }
-
-  content_area = gtk_info_bar_get_content_area (GTK_INFO_BAR (infobar));
-  gtk_container_add (GTK_CONTAINER (content_area), hbox_content);
+  gtk_label_set_text (GTK_LABEL (submit_label), text);
+  g_free (text);
 }
 
-/* Taken from gedit */
 static void
-set_message_area (GtkWidget *container,
-                  GtkWidget *message_area)
+submit_tracks_enabled_changed_cb (GActionGroup *action_group,
+                                  gchar        *action_name,
+                                  gboolean      enabled,
+                                  gpointer      user_data)
 {
-  if (current_message_area == message_area)
-    return;
-
-  if (current_message_area != NULL)
-    gtk_widget_destroy (current_message_area);
-
-  current_message_area = message_area;
-
-  if (message_area == NULL)
-    return;
-
-  gtk_container_add (GTK_CONTAINER (container), message_area);
-
-  g_object_add_weak_pointer (G_OBJECT (current_message_area),
-                             (gpointer)&current_message_area);
+  if (!enabled)
+    gtk_widget_hide (submit_bar);
 }
 
-static GtkWidget*
-musicbrainz_submit_info_bar_new (char *title, char *artist)
+G_MODULE_EXPORT void
+submit_bar_response_cb (GtkInfoBar *infobar,
+                        int         response_id,
+                        gpointer    user_data)
 {
-  GtkWidget *infobar, *button;
-  char *primary_text;
-
-  g_return_val_if_fail (title != NULL, NULL);
-  g_return_val_if_fail (artist != NULL, NULL);
-
-  infobar = gtk_info_bar_new ();
-  button = gtk_info_bar_add_button (GTK_INFO_BAR (infobar),
-                                    _("S_ubmit Album"), GTK_RESPONSE_OK);
-  gtk_info_bar_add_button (GTK_INFO_BAR (infobar),
-                           _("Ca_ncel"), GTK_RESPONSE_CANCEL);
-
-  /* Translators: title, artist */
-  primary_text = g_strdup_printf (_("Could not find %s by %s on MusicBrainz."), title, artist);
-
-  set_info_bar_text_and_icon (GTK_INFO_BAR (infobar),
-                              "dialog-information",
-                              primary_text,
-                              _("You can improve the MusicBrainz database by adding this album."),
-                              button);
-
-  g_free (primary_text);
-
-  return infobar;
+  gtk_widget_hide (GTK_WIDGET (infobar));
 }
 
 /**
@@ -454,6 +388,7 @@ static void on_submit_activate (GSimpleAction *action, GVariant *parameter, gpoi
       g_error_free (error);
     }
   }
+  gtk_widget_hide (GTK_WIDGET (submit_bar));
 }
 
 static void on_preferences_activate (GSimpleAction *action, GVariant *parameter, gpointer data)
@@ -493,18 +428,6 @@ static void on_duplicate_activate (GSimpleAction *action, GVariant *parameter, g
   }
 }
 
-static void
-musicbrainz_submit_info_bar_response (GtkInfoBar *infobar,
-                                      int         response_id,
-                                      gpointer    user_data)
-{
-  if (response_id == GTK_RESPONSE_OK) {
-    on_submit_activate (NULL, NULL, NULL);
-  }
-
-  set_message_area (message_area_eventbox, NULL);
-}
-
 #define TABLE_ROW_SPACING 6 /* spacing of rows in entry_table */
 
 /*
@@ -629,8 +552,6 @@ static void update_ui_for_album (AlbumDetails *album)
     set_action_enabled ("previous-track", FALSE);
     set_action_enabled ("next-track", FALSE);
     set_action_enabled ("duplicate", FALSE);
-
-    set_message_area (message_area_eventbox, NULL);
   } else {
     gtk_list_store_clear (track_store);
 
@@ -709,24 +630,10 @@ static void update_ui_for_album (AlbumDetails *album)
     } else {
       gtk_label_set_text (GTK_LABEL (duration_label), _("(unknown)"));
     }
-
     /* If album details don't come from MusicBrainz ask user to add them */
     if (album->metadata_source != SOURCE_MUSICBRAINZ) {
-      GtkWidget *infobar;
-
-      infobar = musicbrainz_submit_info_bar_new (album->title, album->artist);
-
-      set_message_area (message_area_eventbox, infobar);
-
-      g_signal_connect (infobar,
-                        "response",
-                        G_CALLBACK (musicbrainz_submit_info_bar_response),
-                        NULL);
-
-      gtk_info_bar_set_default_response (GTK_INFO_BAR (infobar),
-                                         GTK_RESPONSE_CANCEL);
-
-      gtk_widget_show (infobar);
+      set_submit_text (album);
+      gtk_widget_show (submit_bar);
     }
   }
 }
@@ -1212,6 +1119,8 @@ metadata_cb (SjMetadataGetter *m, GList *albums, GError *error)
 {
   gboolean realized = gtk_widget_get_realized (main_window);
 
+  set_action_state ("re-read(false)");
+
   if (realized)
     gdk_window_set_cursor (gtk_widget_get_window (main_window), NULL);
     /* Clear the statusbar message */
@@ -1305,6 +1214,7 @@ static void reread_cd (gboolean ignore_no_media)
 
   window = gtk_widget_get_window (main_window);
 
+  set_action_state ("re-read(true)");
   set_action_enabled ("re-read", FALSE);
 
   /* Make sure nothing is playing */
@@ -1330,6 +1240,7 @@ static void reread_cd (gboolean ignore_no_media)
 
   if (!is_audio_cd (drive)) {
     sj_debug (DEBUG_CD, "Media is not an audio CD\n");
+    set_action_state ("re-read(false)");
     update_ui_for_album (NULL);
     gtk_statusbar_pop(GTK_STATUSBAR(status_bar), 0);
     if (realized)
@@ -1370,7 +1281,9 @@ media_added_cb (BraseroMediumMonitor	*drive,
   }
 
   sj_debug (DEBUG_CD, "Media added to device %s\n", brasero_drive_get_device (brasero_medium_get_drive (medium)));
-  reread_cd (TRUE);
+  /* Don't call re-read if metadata is already being retreived */
+  if (!get_action_state_bool ("re-read"))
+    reread_cd (TRUE);
 }
 
 static void
@@ -1388,6 +1301,7 @@ media_removed_cb (BraseroMediumMonitor	*drive,
   sj_debug (DEBUG_CD, "Media removed from device %s\n", brasero_drive_get_device (brasero_medium_get_drive (medium)));
   stop_ui_hack ();
   update_ui_for_album (NULL);
+  set_action_enabled ("submit-tracks", FALSE);
 }
 
 static void
@@ -1574,10 +1488,11 @@ static void profile_changed_cb (GSettings *settings, gchar *key, gpointer user_d
                                      GTK_MESSAGE_QUESTION,
                                      GTK_BUTTONS_NONE,
                                      _("The currently selected audio profile is not available on your installation."));
-    gtk_dialog_add_button (GTK_DIALOG (dialog), "gtk-quit", GTK_RESPONSE_REJECT);
+    gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Quit"), GTK_RESPONSE_REJECT);
     gtk_dialog_add_button (GTK_DIALOG (dialog), _("_Change Profile"), GTK_RESPONSE_ACCEPT);
+    gtk_dialog_set_default_response (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT);
     response = gtk_dialog_run (GTK_DIALOG (dialog));
-    if (response == GTK_RESPONSE_ACCEPT) {
+    if (response != GTK_RESPONSE_REJECT) {
       gtk_widget_destroy (dialog);
       show_preferences_dialog ();
     } else {
@@ -1599,55 +1514,100 @@ static void on_reread_activate (GSimpleAction *action, GVariant *parameter, gpoi
 }
 
 /**
- * Called in on_extract_toggled to see if there are selected tracks or not.
- * extracting points to the boolean to set to if there are tracks to extract,
- * and starts as false.
+ * Move focus to next widget when GtkEntry is activated
  */
-static gboolean extract_available_foreach (GtkTreeModel *model,
-                                           GtkTreePath *path,
-                                           GtkTreeIter *iter,
-                                           gboolean *extracting)
+G_MODULE_EXPORT void on_activate_move_focus (GtkWidget *widget,
+                                             gpointer   data)
 {
-  gboolean selected;
-  gtk_tree_model_get (GTK_TREE_MODEL (track_store), iter, COLUMN_EXTRACT, &selected, -1);
-  if (selected) {
-    *extracting = TRUE;
-    /* Don't bother walking the list more, we've found a track to be extracted. */
-    return TRUE;
+  gboolean ret_val;
+
+  g_signal_emit_by_name (widget, "move-focus", GTK_DIR_TAB_FORWARD, &ret_val);
+}
+
+/**
+ * If path is selected call func on all selected rows, if not call
+ * func only on path.
+ */
+static void
+tree_path_or_selection_foreach (GtkTreeView                 *tree_view,
+                                GtkTreePath                 *path,
+                                GtkTreeSelectionForeachFunc  func,
+                                gpointer                     data)
+{
+  GtkTreeSelection *selection;
+
+  selection = gtk_tree_view_get_selection (tree_view);
+  if (gtk_tree_selection_path_is_selected (selection, path)) {
+    gtk_tree_selection_selected_foreach (selection,
+                                         func,
+                                         data);
   } else {
-    return FALSE;
+    GtkTreeModel *model;
+    GtkTreeIter iter;
+    model = gtk_tree_view_get_model (tree_view);
+    if (gtk_tree_model_get_iter (model, &iter, path))
+      func (model, path, &iter, data);
+  }
+}
+
+/**
+ * Called by on_extract_toggled to update the Extract column for all
+ * the selected rows
+ */
+static void
+on_extract_toggled_foreach (GtkTreeModel *model,
+                            GtkTreePath  *path,
+                            GtkTreeIter  *iter,
+                            gpointer      data)
+{
+  gboolean extract = GPOINTER_TO_INT (data);
+  gboolean old_extract;
+
+  gtk_tree_model_get (model, iter,
+                      COLUMN_EXTRACT, &old_extract,
+                      -1);
+  if (extract != old_extract) {
+    gtk_list_store_set (GTK_LIST_STORE (model), iter,
+                        COLUMN_EXTRACT, extract,
+                        -1);
+    /* Update number of selected tracks */
+    if (extract) {
+      no_of_tracks_selected++;
+    } else {
+      no_of_tracks_selected--;
+    }
   }
 }
 
 /**
  * Called when the user clicked on the Extract column check boxes
  */
-static void on_extract_toggled (GtkCellRendererToggle *cellrenderertoggle,
-                                gchar *path,
-                                gpointer user_data)
+static void
+on_extract_toggled (GtkCellRendererToggle *cellrenderertoggle,
+                    gchar                 *path_str,
+                    gpointer               user_data)
 {
   gboolean extract;
+  GtkTreePath *path;
   GtkTreeIter iter;
 
-  if (!gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (track_store), &iter, path))
+  if (!gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (track_store),
+                                            &iter,
+                                            path_str))
       return;
-  gtk_tree_model_get (GTK_TREE_MODEL (track_store), &iter, COLUMN_EXTRACT, &extract, -1);
+  gtk_tree_model_get (GTK_TREE_MODEL (track_store), &iter,
+                      COLUMN_EXTRACT, &extract,
+                      -1);
   /* extract is the old state here, so toggle */
   extract = !extract;
-  gtk_list_store_set (track_store, &iter, COLUMN_EXTRACT, extract, -1);
-
-  /* Update the Extract buttons */
-  if (extract) {
-    /* If true, then we can extract */
-    gtk_widget_set_sensitive (extract_button, TRUE);
-    no_of_tracks_selected++;
-  } else {
-    /* Reuse the boolean extract */
-    extract = FALSE;
-    gtk_tree_model_foreach (GTK_TREE_MODEL (track_store), (GtkTreeModelForeachFunc)extract_available_foreach, &extract);
-    gtk_widget_set_sensitive (extract_button, extract);
-    no_of_tracks_selected--;
-  }
+  path = gtk_tree_path_new_from_string (path_str);
+  tree_path_or_selection_foreach (GTK_TREE_VIEW (track_listview),
+                                  path,
+                                  on_extract_toggled_foreach,
+                                  GINT_TO_POINTER (extract));
+  gtk_tree_path_free (path);
+  /* Update the Extract button */
+  gtk_widget_set_sensitive (extract_button, no_of_tracks_selected > 0);
   /* Enable and disable the Select/Deselect All buttons */
   if (no_of_tracks_selected == total_no_of_tracks) {
     set_action_enabled ("deselect-all", TRUE);
@@ -1661,33 +1621,60 @@ static void on_extract_toggled (GtkCellRendererToggle *cellrenderertoggle,
   }
 }
 
-/**
- * Callback when the title, artist or composer cells are edited in the list.
- * column_data contains the column number in the model which was modified.
- */
-static void on_cell_edited (GtkCellRendererText *renderer,
-                 gchar *path, gchar *string,
-                 gpointer column_data)
+static void
+on_cell_changed_foreach (GtkTreeModel *model,
+                         GtkTreePath  *path,
+                         GtkTreeIter  *iter,
+                         gpointer      user_data)
 {
-  ViewColumn column = GPOINTER_TO_INT (column_data);
-  GtkTreeIter iter;
+  CellCbContext *context = user_data;
+
+  /* Updating the model for the cell being edited aborts the edit so
+     avoid it and set it in on_cell_editing_done instead */
+  if (gtk_tree_path_compare (path, context->editable_path) != 0)
+    gtk_list_store_set (GTK_LIST_STORE (model), iter,
+                        context->column, context->text,
+                        -1);
+}
+
+/**
+ * Update the selected rows as the user types
+ */
+static void
+on_cell_changed (GtkEditable *editable,
+                 gpointer     user_data)
+{
+  CellCbContext *context = user_data;
+
+  context->text = gtk_editable_get_chars (editable, 0, -1);
+  tree_path_or_selection_foreach (GTK_TREE_VIEW (track_listview),
+                                  context->editable_path,
+                                  on_cell_changed_foreach,
+                                  context);
+  g_free (context->text);
+}
+
+static void
+on_cell_editing_done_foreach (GtkTreeModel *model,
+                              GtkTreePath  *path,
+                              GtkTreeIter  *iter,
+                              gpointer      user_data)
+{
+  CellCbContext *context = user_data;
   TrackDetails *track;
 
-  if (!gtk_tree_model_get_iter_from_string (GTK_TREE_MODEL (track_store), &iter, path))
-    return;
-  gtk_tree_model_get (GTK_TREE_MODEL (track_store), &iter,
+  /* COLUMN_DETAILS isn't updated during editing so update it now */
+  gtk_tree_model_get (GTK_TREE_MODEL (track_store), iter,
                       COLUMN_DETAILS, &track,
                       -1);
-  switch (column) {
+  switch (context->column) {
   case COLUMN_TITLE:
     g_free (track->title);
-    track->title = g_strdup (string);
-    gtk_list_store_set (track_store, &iter, COLUMN_TITLE, track->title, -1);
+    track->title = g_strdup (context->text);
     break;
   case COLUMN_ARTIST:
     g_free (track->artist);
-    track->artist = g_strdup (string);
-    gtk_list_store_set (track_store, &iter, COLUMN_ARTIST, track->artist, -1);
+    track->artist = g_strdup (context->text);
     if (track->artist_sortname) {
       g_free (track->artist_sortname);
       track->artist_sortname = NULL;
@@ -1699,18 +1686,101 @@ static void on_cell_edited (GtkCellRendererText *renderer,
     break;
   case COLUMN_COMPOSER:
     g_free (track->composer);
-    track->composer = g_strdup (string);
-    gtk_list_store_set (track_store, &iter, COLUMN_COMPOSER, track->composer, -1);
+    track->composer = g_strdup (context->text);
     if (track->composer_sortname) {
       g_free (track->composer_sortname);
       track->composer_sortname = NULL;
     }
     break;
   default:
-    g_warning (_("Unknown column %d was edited"), column);
+    g_warning ("Unknown column %d in on_cell_editing_done_foreach",
+               context->column);
   }
+}
 
-  return;
+static void
+on_cell_editing_canceled_foreach (GtkTreeModel *model,
+                                  GtkTreePath  *path,
+                                  GtkTreeIter  *iter,
+                                  gpointer      user_data)
+{
+  CellCbContext *context = user_data;
+  TrackDetails *track;
+  char *text;
+
+  /* COLUMN_DETAILS isn't updated during editing so get original
+     values back from it. */
+  gtk_tree_model_get (model, iter, COLUMN_DETAILS, &track, -1);
+  switch (context->column) {
+  case COLUMN_TITLE:
+    text = track->title;
+    break;
+  case COLUMN_ARTIST:
+    text = track->artist;
+    break;
+  case COLUMN_COMPOSER:
+    text = track->composer;
+    break;
+  default:
+    g_warning ("Unknown column %d in on_cell_editing_canceled_foreach",
+               context->column);
+    return;
+  }
+  gtk_list_store_set (track_store, iter, context->column, text, -1);
+}
+
+/**
+ *  Called when the user finishes or cancels editing a track. Store
+ *  changes/restore original values & disconnect callbacks on the
+ *  GtkCellEditable.
+ */
+static void
+on_cell_editing_done (GtkCellEditable *cell_editable,
+                      gpointer         user_data)
+{
+  gboolean canceled;
+  CellCbContext *context = user_data;
+
+  g_signal_handlers_disconnect_by_data (cell_editable, context);
+  g_object_get (G_OBJECT (cell_editable), "editing-canceled", &canceled, NULL);
+  if (canceled) {
+    tree_path_or_selection_foreach (GTK_TREE_VIEW (track_listview),
+                                    context->editable_path,
+                                    on_cell_editing_canceled_foreach,
+                                    context);
+  } else { /* We need to update the track details in COLUMN_DETAILS */
+    GtkTreeIter iter;
+    context->text = gtk_editable_get_chars (GTK_EDITABLE (cell_editable), 0, -1);
+    /* Update the model for the cell that has been edited */
+    gtk_tree_model_get_iter (GTK_TREE_MODEL (track_store),
+                             &iter, context->editable_path);
+    gtk_list_store_set (track_store, &iter, context->column, context->text, -1);
+    /* Update COLUMN_DETAILS */
+    tree_path_or_selection_foreach (GTK_TREE_VIEW (track_listview),
+                                    context->editable_path,
+                                    on_cell_editing_done_foreach,
+                                    context);
+    g_free (context->text);
+  }
+  gtk_tree_path_free (context->editable_path);
+}
+
+/**
+ * Called when the user starts editing a track. Setup callbacks on the
+ * GtkCellEditable to support editing multiple rows at once.
+ */
+static void
+on_cell_editing_started (GtkCellRenderer *renderer,
+                         GtkCellEditable *cell_editable,
+                         gchar           *path,
+                         gpointer         user_data)
+{
+  cell_editing_context.editable_path = gtk_tree_path_new_from_string (path);
+  cell_editing_context.column = GPOINTER_TO_UINT (user_data);
+  g_signal_connect (cell_editable, "changed",
+                    G_CALLBACK (on_cell_changed), &cell_editing_context);
+  g_signal_connect (cell_editable, "editing-done",
+                    G_CALLBACK (on_cell_editing_done), &cell_editing_context);
 }
 
 /*
@@ -1964,7 +2034,7 @@ is_cd_duplication_available(void)
 }
 
 GActionEntry app_entries[] = {
-  { "re-read", on_reread_activate, NULL, NULL, NULL },
+  { "re-read", on_reread_activate, NULL, "false", NULL },
   { "duplicate", on_duplicate_activate, NULL, NULL, NULL },
   { "eject", on_eject_activate, NULL, NULL, NULL },
   { "submit-tracks", on_submit_activate, NULL, NULL, NULL },
@@ -1982,6 +2052,28 @@ GActionEntry win_entries[] = {
   { "deselect-all", on_deselect_all_activate, NULL, NULL, NULL }
 };
 
+static GtkCellRenderer*
+add_editable_listview_column (const char       *title,
+                              const ViewColumn  model_column)
+{
+  GtkCellRenderer *renderer;
+  GtkTreeViewColumn *column;
+
+  renderer = sj_cell_renderer_text_new ();
+  g_signal_connect (renderer,
+                    "editing-started",
+                    G_CALLBACK (on_cell_editing_started),
+                    GUINT_TO_POINTER (model_column));
+  g_object_set (G_OBJECT (renderer), "editable", TRUE, NULL);
+  column = gtk_tree_view_column_new_with_attributes (title, renderer,
+                                                     "text", model_column,
+                                                     NULL);
+  gtk_tree_view_column_set_resizable (column, TRUE);
+  gtk_tree_view_column_set_expand (column, TRUE);
+  gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
+  return renderer;
+}
+
 static void
 startup_cb (GApplication *app, gpointer user_data)
 {
@@ -2045,7 +2137,8 @@ startup_cb (GApplication *app, gpointer user_data)
   gtk_builder_connect_signals (builder, NULL);
 
   main_window           = GET_WIDGET ("main_window");
-  message_area_eventbox = GET_WIDGET ("message_area_eventbox");
+  submit_bar            = GET_WIDGET ("submit_bar");
+  submit_label          = GET_WIDGET ("submit_label");
   title_entry           = GET_WIDGET ("title_entry");
   artist_entry          = GET_WIDGET ("artist_entry");
   composer_label        = GET_WIDGET ("composer_label");
@@ -2066,6 +2159,10 @@ startup_cb (GApplication *app, gpointer user_data)
   g_action_map_add_action_entries (G_ACTION_MAP (main_window),
                                    win_entries, G_N_ELEMENTS (win_entries),
                                    NULL);
+
+  g_signal_connect (app, "action-enabled-changed::submit-tracks",
+                    G_CALLBACK (submit_tracks_enabled_changed_cb), NULL);
+
   gtk_button_set_label(GTK_BUTTON(select_button), _("Select None"));
   gtk_actionable_set_action_name(GTK_ACTIONABLE(select_button), "win.deselect-all");
   gtk_actionable_set_action_name(GTK_ACTIONABLE(play_button), "win.play");
@@ -2185,40 +2282,12 @@ startup_cb (GApplication *app, gpointer user_data)
     gtk_tree_view_column_set_cell_data_func (column, renderer, number_cell_icon_data_cb, NULL, NULL);
     gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
 
-    title_renderer = gtk_cell_renderer_text_new ();
-    g_signal_connect (title_renderer, "edited", G_CALLBACK (on_cell_edited), GUINT_TO_POINTER (COLUMN_TITLE));
-    g_object_set (G_OBJECT (title_renderer), "editable", TRUE, NULL);
-    column = gtk_tree_view_column_new_with_attributes (_("Title"),
-                                                       title_renderer,
-                                                       "text", COLUMN_TITLE,
-                                                       NULL);
-    gtk_tree_view_column_set_resizable (column, TRUE);
-    gtk_tree_view_column_set_expand (column, TRUE);
-    gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
-
-    artist_renderer = gtk_cell_renderer_text_new ();
-    column = gtk_tree_view_column_new_with_attributes (_("Artist"),
-                                                       artist_renderer,
-                                                       "text", COLUMN_ARTIST,
-                                                       NULL);
-    gtk_tree_view_column_set_resizable (column, TRUE);
-    gtk_tree_view_column_set_expand (column, TRUE);
-    g_signal_connect (artist_renderer, "edited", G_CALLBACK (on_cell_edited), GUINT_TO_POINTER (COLUMN_ARTIST));
-    g_object_set (G_OBJECT (artist_renderer), "editable", TRUE, NULL);
-    gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
-
-    composer_renderer = gtk_cell_renderer_text_new ();
-    column = gtk_tree_view_column_new_with_attributes (_("Composer"),
-                                                       composer_renderer,
-                                                       "text", COLUMN_COMPOSER,
-                                                       NULL);
-    gtk_tree_view_column_set_resizable (column, TRUE);
-    gtk_tree_view_column_set_expand (column, TRUE);
-    g_signal_connect (composer_renderer, "edited", G_CALLBACK (on_cell_edited), GUINT_TO_POINTER (COLUMN_COMPOSER));
-    g_object_set (G_OBJECT (composer_renderer), "editable", TRUE, NULL);
-    gtk_tree_view_column_set_visible (column, FALSE);
-    composer_column = column;
-    gtk_tree_view_append_column (GTK_TREE_VIEW (track_listview), column);
+    title_renderer = add_editable_listview_column (_("Title"), COLUMN_TITLE);
+    artist_renderer = add_editable_listview_column (_("Artist"), COLUMN_ARTIST);
+    composer_renderer = add_editable_listview_column (_("Composer"), COLUMN_COMPOSER);
+    composer_column = gtk_tree_view_get_column (GTK_TREE_VIEW (track_listview),
+                                                gtk_tree_view_get_n_columns (GTK_TREE_VIEW (track_listview)) - 1);
+    gtk_tree_view_column_set_visible (composer_column, FALSE);
 
     renderer = gtk_cell_renderer_text_new ();
     column = gtk_tree_view_column_new_with_attributes (_("Duration"),
@@ -2241,7 +2310,7 @@ startup_cb (GApplication *app, gpointer user_data)
   sj_play_init ();
 
   selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (track_listview));
-  gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
+  gtk_tree_selection_set_mode (selection, GTK_SELECTION_MULTIPLE);
 
   baseuri_changed_cb (sj_settings, SJ_SETTINGS_BASEURI, NULL);
   path_pattern_changed_cb (sj_settings, SJ_SETTINGS_PATH_PATTERN, NULL);
@@ -2299,15 +2368,68 @@ activate_cb (GApplication *app, gpointer user_data)
   gtk_window_present (GTK_WINDOW (main_window));
 }
 
-void set_action_enabled (const char *name, gboolean enabled)
+static GAction*
+lookup_action (const gchar* name)
 {
-  GActionMap *map = G_ACTION_MAP (g_application_get_default ());
-  GAction *action = g_action_map_lookup_action (map, name);
+  GActionMap *map;
+  GAction *action;
 
+  map = G_ACTION_MAP (g_application_get_default ());
+  action = g_action_map_lookup_action (map, name);
   if (action == NULL)
     action = g_action_map_lookup_action (G_ACTION_MAP (main_window), name);
+  if (action == NULL)
+    g_warning ("Action '%s' not found", name);
 
-  g_simple_action_set_enabled (G_SIMPLE_ACTION (action), enabled);
+  return action;
+}
+
+static gboolean
+get_action_state_bool (const gchar* name)
+{
+  gboolean state = FALSE;
+  GAction *action;
+
+  action = lookup_action (name);
+  if (action != NULL) {
+    GVariant *value = g_action_get_state (action);
+    state = g_variant_get_boolean (value);
+    g_variant_unref (value);
+  }
+
+  return state;
+}
+
+static void
+set_action_state (const gchar *detailed_name)
+{
+  gchar *name;
+  GVariant *value;
+  GError *error = NULL;
+
+  if (g_action_parse_detailed_name (detailed_name, &name, &value, &error)) {
+    GSimpleAction *action;
+
+    action = G_SIMPLE_ACTION (lookup_action (name));
+    if (action != NULL)
+      g_simple_action_set_state (action, value);
+
+    g_free (name);
+    g_variant_unref (value);
+  } else {
+    g_warning ("Error parsing detailed action name '%s' - %s",
+               detailed_name, error->message);
+    g_error_free (error);
+  }
+}
+
+void set_action_enabled (const char *name, gboolean enabled)
+{
+  GAction *action;
+
+  action = lookup_action (name);
+  if (action != NULL)
+    g_simple_action_set_enabled (G_SIMPLE_ACTION (action), enabled);
 }
 
 int main (int argc, char **argv)
